from __future__ import annotations
from copy import deepcopy

from .vec3 import Vec3


class Mat3:
    def __init__(self, mat: mat_type) -> None:

        if not is_mat_type(mat):
            raise TypeError(f"Expected 3x3 matrix list, got {type(mat)}")
        self.mat = [Vec3(*row) for row in mat]

    def __str__(self) -> str:
        return f"{str(self.mat[0])}\n{str(self.mat[1])}\n{str(self.mat[2])}"

    def __repr__(self) -> str:
        return f"Mat3({repr(self.mat[0])}\n{repr(self.mat[1])}\n{repr(self.mat[2])})"

    def __eq__(self, other: Mat3) -> bool:
        return self.mat == other.mat

    def __ne__(self, other: Mat3) -> bool:
        return not self == other

    def __getitem__(self, index: int) -> Vec3:
        return self.mat[index]

    def __setitem__(self, index: int, value: vec_type) -> None:
        if not is_vec_type(value):
            raise TypeError(f"Expected 3 len vector, got {type(value)}")
        if isinstance(value, Vec3):
            self.mat[index] = value
        else:
            self.mat[index] = Vec3(*value)

    def __iter__(self) -> Vec3:
        for i in range(3):
            yield Vec3(*self.mat[i])

    def __len__(self) -> int:
        return 3

    def __neg__(self) -> Mat3:
        return Mat3([[-self.mat[i][j] for j in range(3)] for i in range(3)])

    def __add__(self, other: Mat3) -> Mat3:
        if not isinstance(other, Mat3):
            raise TypeError(f"Expected 3x3 matrix, got {type(other)}")
        return Mat3(
            [[self.mat[i][j] + other.mat[i][j] for j in range(3)] for i in range(3)]
        )

    def __radd__(self, other: Mat3) -> Mat3:
        if not isinstance(other, Mat3):
            raise TypeError(f"Expected 3x3 matrix, got {type(other)}")
        return Mat3(
            [[self.mat[i][j] + other.mat[i][j] for j in range(3)] for i in range(3)]
        )

    def __iadd__(self, other: Mat3) -> Mat3:
        if not isinstance(other, Mat3):
            raise TypeError(f"Expected 3x3 matrix, got {type(other)}")
        for i in range(3):
            for j in range(3):
                self.mat[i][j] += other.mat[i][j]
        return self

    def __sub__(self, other: Mat3) -> Mat3:
        if not isinstance(other, Mat3):
            raise TypeError(f"Expected 3x3 matrix, got {type(other)}")
        return Mat3(
            [[self.mat[i][j] - other.mat[i][j] for j in range(3)] for i in range(3)]
        )

    def __rsub__(self, other: Mat3) -> Mat3:
        if not isinstance(other, Mat3):
            raise TypeError(f"Expected 3x3 matrix, got {type(other)}")
        return Mat3(
            [[self.mat[i][j] - other.mat[i][j] for j in range(3)] for i in range(3)]
        )

    def __isub__(self, other: Mat3) -> Mat3:
        if not isinstance(other, Mat3):
            raise TypeError(f"Expected 3x3 matrix, got {type(other)}")
        for i in range(3):
            for j in range(3):
                self.mat[i][j] -= other.mat[i][j]
        return self

    def __mul__(self, other: int | float) -> Mat3:
        if not isinstance(other, (int, float)):
            raise TypeError(f"Expected scalar, got {type(other)}")
        return Mat3([[self.mat[i][j] * other for j in range(3)] for i in range(3)])

    def __rmul__(self, other: int | float) -> Mat3:
        if not isinstance(other, (int, float)):
            raise TypeError(f"Expected scalar, got {type(other)}")
        return Mat3([[self.mat[i][j] * other for j in range(3)] for i in range(3)])

    def __imul__(self, other: int | float) -> Mat3:
        if not isinstance(other, (int, float)):
            raise TypeError(f"Expected scalar, got {type(other)}")
        for i in range(3):
            for j in range(3):
                self.mat[i][j] *= other
        return self

    def __matmul__(self, other: Mat3 | Vec3) -> Mat3 | Vec3:
        if not isinstance(other, (Mat3, Vec3)):
            raise TypeError(f"Expected Mat3 or Vec3, got {type(other)}")

        if isinstance(other, Vec3):
            return Vec3.from_list(
                [sum(self.mat[i][j] * other[j] for j in range(3)) for i in range(3)]
            )
        else:
            return Mat3(
                [
                    [
                        sum(self.mat[i][k] * other.mat[k][j] for k in range(3))
                        for j in range(3)
                    ]
                    for i in range(3)
                ]
            )

    def __rmatmul__(self, other: Mat3 | Vec3) -> Mat3 | Vec3:
        if not isinstance(other, (Mat3, Vec3)):
            raise TypeError(f"Expected Mat3 or Vec3, got {type(other)}")

        if isinstance(other, Vec3):
            return Vec3.from_list(
                [sum(self.mat[i][j] * other[j] for j in range(3)) for i in range(3)]
            )
        else:
            return Mat3(
                [
                    [
                        sum(self.mat[i][k] * other.mat[k][j] for k in range(3))
                        for j in range(3)
                    ]
                    for i in range(3)
                ]
            )

    def __imatmul__(self, other: Mat3) -> Mat3:
        if not isinstance(other, Mat3):
            raise TypeError(f"Expected 3x3 matrix, got {type(other)}")
        for i in range(3):
            for j in range(3):
                self.mat[i][j] = sum(self.mat[i][k] * other.mat[k][j] for k in range(3))
        return self

    def __truediv__(self, other: int | float) -> Mat3:
        if not isinstance(other, (int, float)):
            raise TypeError(f"Expected scalar, got {type(other)}")
        return Mat3([[self.mat[i][j] / other for j in range(3)] for i in range(3)])

    def __itruediv__(self, other: int | float) -> Mat3:
        if not isinstance(other, (int, float)):
            raise TypeError(f"Expected scalar, got {type(other)}")
        for i in range(3):
            for j in range(3):
                self.mat[i][j] /= other
        return self

    def __floordiv__(self, other: int | float) -> Mat3:
        if not isinstance(other, (int, float)):
            raise TypeError(f"Expected scalar, got {type(other)}")
        return Mat3([[self.mat[i][j] // other for j in range(3)] for i in range(3)])

    def __ifloordiv__(self, other: int | float) -> Mat3:
        if not isinstance(other, (int, float)):
            raise TypeError(f"Expected scalar, got {type(other)}")
        for i in range(3):
            for j in range(3):
                self.mat[i][j] //= other
        return self

    def __mod__(self, other: int | float) -> Mat3:
        if not isinstance(other, (int, float)):
            raise TypeError(f"Expected scalar, got {type(other)}")
        return Mat3([[self.mat[i][j] % other for j in range(3)] for i in range(3)])

    def __imod__(self, other: int | float) -> Mat3:
        if not isinstance(other, (int, float)):
            raise TypeError(f"Expected scalar, got {type(other)}")
        for i in range(3):
            for j in range(3):
                self.mat[i][j] %= other
        return self

    def __pow__(self, other: int | float) -> Mat3:
        if not isinstance(other, (int, float)):
            raise TypeError(f"Expected scalar, got {type(other)}")
        return Mat3([[self.mat[i][j] ** other for j in range(3)] for i in range(3)])

    def __ipow__(self, other: int | float) -> Mat3:
        if not isinstance(other, (int, float)):
            raise TypeError(f"Expected scalar, got {type(other)}")
        for i in range(3):
            for j in range(3):
                self.mat[i][j] **= other
        return self

    def copy(self) -> Mat3:
        return Mat3(deepcopy(self.mat))

    def to_list(self) -> list[list[float]]:
        return deepcopy(self.mat)

    def matrix_power(self, other: int | float) -> Mat3:
        """calculate the power of the matrix, self**3 = self*self*self"""
        if not isinstance(other, (int, float)):
            raise TypeError(f"Expected scalar, got {type(other)}")
        if other == 0:
            return Mat3([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
        if other < 0:
            raise ValueError("minus power not supported yet")
        tmp = self.copy()
        for _ in range(other - 1):
            tmp = tmp @ self
        return tmp

    def T(self) -> Mat3:
        return Mat3([[self.mat[j][i] for j in range(3)] for i in range(3)])

    def det(self) -> float:
        return (
            self[0][0] * self[1][1] * self[2][2]
            + self[0][1] * self[1][2] * self[2][0]
            + self[0][2] * self[1][0] * self[2][1]
            - self[0][2] * self[1][1] * self[2][0]
            - self[0][0] * self[1][2] * self[2][1]
            - self[0][1] * self[1][0] * self[2][2]
        )

    def inv(self) -> Mat3:
        """
        mat^(-1) = 1/det(mat) * [[a22a33 - a23a32, -(a12a33 - a13a32), a12a23 - a13a22],
                                 [-(a21a33 - a23a31), a11a33 - a13a31, -(a11a23 - a13a21)],
                                 [a21a32 - a22a31, -(a11a32 - a12a31), a11a22 - a12a21]]
        """
        det = self.det()
        if det == 0:
            raise ValueError("Matrix is not invertible")
        return Mat3(
            [
                [
                    (self[1][1] * self[2][2] - self[1][2] * self[2][1]) / det,
                    -(self[0][1] * self[2][2] - self[0][2] * self[2][1]) / det,
                    (self[0][1] * self[1][2] - self[0][2] * self[1][1]) / det,
                ],
                [
                    -(self[1][0] * self[2][2] - self[1][2] * self[2][0]) / det,
                    (self[0][0] * self[2][2] - self[0][2] * self[2][0]) / det,
                    -(self[0][0] * self[1][2] - self[0][2] * self[1][0]) / det,
                ],
                [
                    (self[1][0] * self[2][1] - self[1][1] * self[2][0]) / det,
                    -(self[0][0] * self[2][1] - self[0][1] * self[2][0]) / det,
                    (self[0][0] * self[1][1] - self[0][1] * self[1][0]) / det,
                ],
            ]
        )


vec_type = Vec3 | list[int | float]
mat_type = list[Vec3 | list[int | float]]


def is_vec_type(vec: list[float]) -> bool:
    return (
        isinstance(vec, (list, Vec3))
        and len(vec) == 3
        and all(isinstance(i, (int, float)) for i in [*vec])
    )


def is_mat_type(mat: list[list[float]] | Mat3) -> bool:
    return isinstance(mat, Mat3) or (
        isinstance(mat, list)
        and len(mat) == 3
        and all(is_vec_type(vec) for vec in mat if isinstance(vec, list))
    )



from molgeom import Atom, Vec3, Mat3


def test_instanciation():
    atom = Atom("C", 0, 1.1, -0.2)

    assert atom.symbol == "C"
    assert atom.x == 0
    assert atom.y == 1.1
    assert atom.z == -0.2


def test_from_vec():
    list_vec = [2, 1.1, -0.2]
    tuple_vec = (2, 1.1, -0.2)
    vec3 = Vec3(2, 1.1, -0.2)
    a1 = Atom.from_vec("C", list_vec)
    a2 = Atom.from_vec("C", tuple_vec)
    a3 = Atom.from_vec("C", vec3)

    assert isinstance(a1, Atom) and isinstance(a2, Atom) and isinstance(a3, Atom)
    assert a1 == a2 == a3
    assert (
        a1.x == a2.x == a3.x == 2
        and a1.y == a2.y == a3.y == 1.1
        and a1.z == a2.z == a3.z == -0.2
    )


def test_to_Vec3():
    atom = Atom("C", 0, 1.1, -0.2)
    vec = atom.to_Vec3()

    assert vec == Vec3(0, 1.1, -0.2)


def test_copy():
    atom = Atom("C", 0, 1.1, -0.2)
    atom_copied = atom.copy()

    assert atom == atom_copied
    assert atom is not atom_copied


def test_get_frac_coords():
    atom = Atom("C", 0.3, 1.1, -0.2)
    lattice_vec = Mat3([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    frac_coords = atom.get_frac_coords(lattice_vec, wrap=False)
    assert frac_coords.isclose(Vec3(0.3, 1.1, -0.2))
    frac_coords = atom.get_frac_coords(lattice_vec, wrap=True)
    assert frac_coords.isclose(Vec3(0.3, 0.1, 0.8))

    lattice_vec = Mat3([[2, 0, 0], [0, 2, 0], [0, 0, 2]])
    frac_coords = atom.get_frac_coords(lattice_vec)
    assert frac_coords.isclose(Vec3(0.15, 0.55, -0.1))
    frac_coords = atom.get_frac_coords(lattice_vec, wrap=True)
    assert frac_coords.isclose(Vec3(0.15, 0.55, 0.9))

    lattice_vec = Mat3([[1, 0, 0], [0, 2, 0], [0, 0, 2]])
    frac_coords = atom.get_frac_coords(lattice_vec)
    assert frac_coords.isclose(Vec3(0.3, 0.55, -0.1))



from collections import UserList
from typing import Any


class FancyIndexingList(UserList):
    def __getitem__(self, key: Any):
        if isinstance(key, (list, tuple)):
            return FancyIndexingList([self.data[i] for i in key])
        return super().__getitem__(key)

    def __iter__(self):
        return iter(self.data)

    def __setitem__(self, key: Any, value: Any):
        if isinstance(key, (list, tuple)):
            if len(key) != len(value):
                raise ValueError("Key and value must have the same length")
            for k, v in zip(key, value):
                self.data[k] = v
        else:
            super().__setitem__(key, value)

    def __delitem__(self, key: Any):
        if isinstance(key, (list, tuple)):
            for k in sorted(key, reverse=True):
                del self.data[k]
        else:
            super().__delitem__(key)

    def __repr__(self):
        return f"FancyIndexingList({self.data})"



from __future__ import annotations

import copy
import json
import threading
from importlib.resources import files
from typing import Any

from molgeom.utils.vec3 import Vec3, mat_type
from molgeom.utils.mat3 import Mat3, is_mat_type
from molgeom.utils.lattice_utils import cart2frac
from molgeom.data.consts import ATOMIC_NUMBERS

_pt_data = None
_bond_pair_data = None
_atomic_data_lock = threading.Lock()


def _load_atomic_data():
    """
    Load atomic data from yaml files to _pt_data
    (periodic table data) and _bond_rad_data (bond data)
    This function is thread safe and will only load the data once
    """
    global _pt_data
    global _bond_rad_data
    if _pt_data is None:
        with _atomic_data_lock:
            if _pt_data is None:
                pt_path = files("molgeom.data").joinpath("periodic_table.json")
                with open(pt_path, "r", encoding="utf-8") as f:
                    _pt_data = json.load(f)
                bonds_path = files("molgeom.data").joinpath("bonds_jmol_ob.json")
                with open(bonds_path, "r", encoding="utf-8") as f:
                    _bond_rad_data = json.load(f)
    return _pt_data, _bond_rad_data


def _load_bond_pair_data():
    """
    Load bond pair data from json file to _bond_pair_data
    This function is thread safe and will only load the data once
    """
    global _bond_pair_data
    if _bond_pair_data is None:
        with _atomic_data_lock:
            if _bond_pair_data is None:
                bond_pair_path = files("molgeom.data").joinpath("bond_pairs_len.json")
                with open(bond_pair_path, "r", encoding="utf-8") as f:
                    _bond_pair_data = json.load(f)
    return _bond_pair_data


class Atom(Vec3):
    def __init__(self, symbol: str, x: float, y: float, z: float) -> None:
        super().__init__(x, y, z)
        if symbol not in ATOMIC_NUMBERS:
            raise ValueError(f"Invalid atomic symbol: {symbol}")
        self.symbol: str = symbol
        self._atomic_data, self._std_bond_rad = self.get_atomic_data(self.symbol)
        self._bond_pairs = _load_bond_pair_data()
        self.mass: float = self._atomic_data.get("Atomic mass", 0.0)
        self.atomic_number: int = self._atomic_data.get("Atomic no", 0.0)
        self.charge: int | float | None = None

    @classmethod
    def from_vec(cls, symbol: str, vec: list | tuple | Vec3) -> Atom:
        if (
            isinstance(vec, (list, tuple))
            and len(vec) == 3
            and all(isinstance(i, (int, float)) for i in vec)
        ):
            return cls(symbol, vec[0], vec[1], vec[2])
        if isinstance(vec, Vec3):
            return cls(symbol, vec.x, vec.y, vec.z)
        return cls(symbol, vec.x, vec.y, vec.z)

    @staticmethod
    def get_atomic_data(symbol):
        _pt_data, _bond_rad_data = _load_atomic_data()
        return _pt_data.get(symbol, {}), _bond_rad_data.get(symbol, {})

    def __getitem__(self, index: int) -> Any:
        return (self.symbol, self.x, self.y, self.z)[index]

    def __eq__(self, other: Atom) -> bool:
        return self.symbol == other.symbol and super().__eq__(other)

    def __ne__(self, other: Atom) -> bool:
        return not self == other

    def __lt__(self, other: Atom) -> bool:
        if self.atomic_number != other.atomic_number:
            return self.atomic_number < other.atomic_number
        if self.mass != other.mass:
            return self.mass < other.mass

    def __str__(self) -> str:
        return (
            f"Atom({self.symbol:2s}, {self.x:19.12f}, {self.y:19.12f}, {self.z:19.12f})"
        )

    def __repr__(self) -> str:
        return f"Atom({self.symbol!r}, {self.x:.12f}, {self.y:.12f}, {self.z:.12f})"

    def copy(self) -> Atom:
        return copy.deepcopy(self)

    def get_frac_coords(self, lattice_vecs: mat_type, wrap=False) -> Vec3:
        if not is_mat_type(lattice_vecs):
            raise ValueError("Invalid lattice vector type")
        if not isinstance(lattice_vecs, Mat3):
            lattice_vecs = Mat3(lattice_vecs)

        return cart2frac(self.to_Vec3(), lattice_vecs, wrap=wrap)

    def to_Vec3(self) -> Vec3:
        return Vec3(self.x, self.y, self.z)

    def to_xyz(self) -> str:
        return f"{self.symbol:2s} {self.x:19.12f} {self.y:19.12f} {self.z:19.12f}"

    def to_dict(self) -> dict:
        return {
            "symbol": self.symbol,
            "x": self.x,
            "y": self.y,
            "z": self.z,
            "mass": self.mass,
            "charge": self.charge,
            "atomic_number": self.atomic_number,
        }

    # def is_bonded_to(self, other, tol=0.12, lower_bound=None, upper_bound=None) -> bool:
    # def is_bonded_to(self, other, tol=0.15) -> bool:
    def get_bond_length(self, other, tol=0.15) -> float | None:
        """
        Returns the bond length between two atoms if they are bonded, else None
        args:
            other: Atom
            tol: float
        """
        dist_angst = self.distance_to(other)
        _bond_pair_data = _load_bond_pair_data()
        estimated_bond_len = self._std_bond_rad + other._std_bond_rad
        std_bond_lens = _bond_pair_data.get(self.symbol, {}).get(
            other.symbol, [estimated_bond_len]
        )

        # if lower_bound is not None and upper_bound is not None:
        #     for std_bond_len in std_bond_lens:
        #         if lower_bound <= std_bond_len <= upper_bound:
        #             return True

        # for std_bond_len in std_bond_lens:
        #     if dist_angst <= std_bond_len + tol:
        #         return True
        # return False

        for std_bond_len in std_bond_lens:
            if dist_angst <= std_bond_len + tol:
                return dist_angst

        return None

    def closest_atom(self, mole, symbol=None) -> Atom:
        closest = mole.atoms[0]
        for atom in mole:
            if symbol is not None and atom.symbol != symbol:
                continue
            if self.distance_to(atom) < self.distance_to(closest):
                closest = atom
        return closest



from pathlib import Path

from molgeom.molecule import Molecule
from molgeom.parsers.mol import from_mol_str
from molgeom.parsers.parser_tools import validate_filepath, zopen


def from_sdf_str(content: str) -> list[Molecule]:
    """Parse SDF format string content and return list of Molecule objects.

    SDF format contains multiple MOL format molecules separated by $$$$
    """
    molecules = []
    # Split content at $$$$ to get individual MOL blocks
    mol_blocks = content.split("$$$$")

    for block in mol_blocks:
        block = block.strip()
        if block:  # Skip empty blocks
            try:
                mol = from_mol_str(block)
                molecules.append(mol)
            except ValueError as e:
                raise ValueError(f"Error parsing MOL block: {e}")

    if not molecules:
        raise ValueError("No valid molecules found in SDF content")

    return molecules


def sdf_parser(filepath: str | Path) -> list[Molecule]:
    """Read SDF file and return list of Molecule objects."""
    filepath = validate_filepath(filepath)
    with zopen(filepath, mode="rt", encoding="utf-8") as f:
        content = f.read()
    mol = from_sdf_str(content)
    mol.name = filepath.stem
    return mol



from __future__ import annotations

import re
from pathlib import Path

from molgeom.utils.vec3 import Vec3
from molgeom.utils.mat3 import Mat3
from molgeom.utils.lattice_utils import frac2cart
from molgeom.atom import Atom
from molgeom.data.consts import ATOMIC_MASSES
from molgeom.molecule import Molecule
from molgeom.parsers.parser_tools import validate_filepath, zopen


def poscar_parser(filepath: str | Path) -> Molecule:
    """
    Parse a POSCAR file and return a Molecule object.

    https://www.vasp.at/wiki/index.php/POSCAR#Full_format_specification

    """
    filepath = validate_filepath(filepath)
    if "poscar" not in filepath.stem.lower():
        raise ValueError(f"{filepath} is not a POSCAR file")

    mol = Molecule()
    with zopen(filepath, mode="rt", encoding="utf-8") as file:

        # replace tabs, non-breaking spaces, and multiple spaces with single space
        lines = file.readlines()
        for i in range(len(lines)):
            lines[i] = re.sub(r"[\s\t\xa0]+", " ", lines[i])
        lines_gen = iter(lines)

        # The first line is in principle a comment line
        line1 = next(lines_gen).strip()

        # Scaling factor
        scale = list(map(float, next(lines_gen).split()[:3]))
        if len(scale) not in [1, 3]:
            raise RuntimeError("The number of scaling factors must be 1 or 3.")
        if len(scale) == 3 and any([x <= 0 for x in scale]):
            raise RuntimeError("All three scaling factors must be positive.")

        # Lattice vectors
        lattice_vec_a = Vec3(*map(float, next(lines_gen).split()[:3]))
        lattice_vec_b = Vec3(*map(float, next(lines_gen).split()[:3]))
        lattice_vec_c = Vec3(*map(float, next(lines_gen).split()[:3]))
        lattice_vecs = Mat3([lattice_vec_a, lattice_vec_b, lattice_vec_c])
        mol.lattice_vecs = lattice_vecs

        if len(scale) == 1:
            # Negative scaling factor corresponds to the cell volume.
            scale = scale[0]
            if scale < 0.0:
                scale = (-1.0 * scale / lattice_vecs.det()) ** (1 / 3)
            lattice_vecs = Mat3([scale * vec for vec in lattice_vecs])
        else:
            lattice_vecs = Mat3([scale[i] * vec for i, vec in enumerate(lattice_vecs)])

        # Atom symbols and number of atoms per symbol
        atom_symbols = next(lines_gen).split()
        for symbol in atom_symbols:
            if symbol not in ATOMIC_MASSES:
                raise ValueError(f"Unknown atom symbol {symbol=}")
        num_atoms_per_symbol = list(map(int, next(lines_gen).split()))

        # Selective dynamics (optional line)
        tmp_line = next(lines_gen).strip()
        selective_dynamics = False
        if tmp_line.lower()[0] == "s":
            selective_dynamics = True

        # Direct or Cartesian coordinates
        if not selective_dynamics:
            coord_type = tmp_line.strip().lower()
        else:
            coord_type = next(lines_gen).strip().lower()

        # Read the atomic coordinates
        if coord_type == "direct":
            for i in range(len(atom_symbols)):
                for _ in range(num_atoms_per_symbol[i]):
                    frac_coords = Vec3(*map(float, next(lines_gen).split()[:3]))
                    cart_coords = frac2cart(frac_coords, lattice_vecs)
                    atom = Atom(
                        symbol=atom_symbols[i],
                        x=cart_coords[0],
                        y=cart_coords[1],
                        z=cart_coords[2],
                    )
                    mol.add_atom(atom)
        elif coord_type == "cartesian":
            for i in range(len(atom_symbols)):
                for _ in range(num_atoms_per_symbol[i]):
                    cart_coords = list(map(float, next(lines_gen).split()[:3]))
                    atom = Atom(
                        symbol=atom_symbols[i],
                        x=cart_coords[0],
                        y=cart_coords[1],
                        z=cart_coords[2],
                    )
                    mol.add_atom(atom)
        else:
            raise ValueError(f"Expected 'Direct' or 'Cartesian', got {coord_type=}")

    return mol



import re

from .vec3 import Vec3
from .mat3 import Mat3


# This function is adapted from the `SymmOp.from_rotation_and_translation.from_xyz_str` method in the pymatgen library.
# Original source: https://github.com/materialsproject/pymatgen/blob/da607e86f9ce8aec942067c6c1a4fda6e04915dd/src/pymatgen/core/operations.py#L462
# License: MIT License
# Copyright (c) 2011-2012 MIT & The Regents of the University of California, through Lawrence Berkeley National Laboratory
# Modifications made by [sio-salt], [2024]
def symmop_from_xyz_str(
    xyz_str: str
) -> (Mat3, Vec3):
    """
    Make rotation matrix and translation vector from symmetry operation xyz string.
    Args:
        xyz_str (str): xyz string of symmetry operation.
            e.g.   ‘x, y, z’, ‘-x, -y, z’,
                   '-x, y + 1/2, -z + 1/2', ‘-2y+1/2, 3x+1/2, z-y+1/2’,
    Returns:
        (Mat3, Vec3): Rotation matrix and translation vector in fractional coordinates.
        if you want to convert transvec to cartesian, use molgeom.frac2cart(transvec).
    """

    ops = xyz_str.strip().lower().replace(" ", "").split(",")
    re_rot = re.compile(r"([+-]?)([\d\.]*)/?([\d\.]*)([x-z])")
    re_trans = re.compile(r"([+-]?)([\d\.]+)/?([\d\.]*)(?![x-z])")

    rot_mat = Mat3([[0.0] * 3 for _ in range(3)])
    trans_vec_frac = Vec3(0, 0, 0)
    for i, op in enumerate(ops):

        # make rot mat
        for match in re_rot.finditer(op):
            # match[0] contains the whole match
            # match[n] (n > 0) contains the n-th group surrounded by ()
            factor = -1.0 if match[1] == "-" else 1.0
            if match[2] != "":
                factor *= (
                    float(match[2]) / float(match[3])
                    if match[3] != ""
                    else float(match[2])
                )
            j = ord(match[4]) - 120
            rot_mat[i][j] = factor

        # make trans vec
        for match in re_trans.finditer(op):
            factor = -1 if match[1] == "-" else 1
            num = (
                float(match[2]) / float(match[3]) if match[3] != "" else float(match[2])
            )
            trans_vec_frac[i] = factor * num

    return rot_mat, trans_vec_frac



from pathlib import Path
from collections import deque

from molgeom.atom import Atom
from molgeom.molecule import Molecule
from molgeom.parsers.parser_tools import (
    remove_trailing_empty_lines,
    validate_filepath,
    zopen,
)


def from_mol_str(content: str) -> Molecule:
    """Parse MOL format string content and return Molecule object.

    Handles V2000 format MOL files with atom coordinates and charges.
    Bonds are ignored as per requirements.
    """
    mol = Molecule()
    lines = deque(remove_trailing_empty_lines(content.strip().split("\n")))

    # Skip header (3 lines: ID, program info, comment)
    for _ in range(3):
        lines.popleft()

    # Parse counts line
    counts = lines.popleft()
    try:
        n_atoms = int(counts[0:3])
        n_bonds = int(counts[3:6])
    except (ValueError, IndexError):
        raise ValueError(f"Invalid counts line: {counts}")

    # Parse atom block
    for _ in range(n_atoms):
        line = lines.popleft().strip()
        try:
            # V2000 format: coordinates are 10 char fields, followed by atom symbol
            x = float(line[0:10])
            y = float(line[10:20])
            z = float(line[20:30])
            symbol = line[31:34].strip()

            # Create atom and add it to molecule
            atom = Atom(symbol=symbol, x=x, y=y, z=z)
            mol.add_atom(atom)

        except (ValueError, IndexError) as e:
            raise ValueError(f"Invalid atom line: {line}\nError: {e}")

    return mol


def mol_parser(filepath: str | Path) -> Molecule:
    """Read MOL file and return Molecule object."""
    filepath = validate_filepath(filepath)
    with zopen(filepath, mode="rt", encoding="utf-8") as f:
        content = f.read()
    mol = from_mol_str(content)
    mol.name = filepath.stem
    return mol



from __future__ import annotations

import os
import copy
from pathlib import Path
from collections.abc import Iterable

from cachetools import cachedmethod, LRUCache
from cachetools.keys import hashkey
import networkx as nx

from molgeom.data.consts import ANGST2BOHR_GAU16, ATOMIC_NUMBERS
from molgeom.utils.fancy_indexing_list import FancyIndexingList
from molgeom.utils.vec3 import Vec3, mat_type, vec_type
from molgeom.utils.mat3 import Mat3, is_mat_type
from molgeom.utils.decorators import args_to_list, args_to_set
from molgeom.utils.lattice_utils import cart2frac, frac2cart, lat_vecs_to_lat_params
from molgeom.utils.symmetry_utils import symmop_from_xyz_str
from molgeom.utils.html_utils import view_mol
from molgeom.atom import Atom


default_tol = 0.15


class Molecule:
    """
    A class to represent a molecule.
    """

    def __init__(
        self,
        *atoms: Atom,
        lattice_vecs: mat_type | None = None,
        name: str | None = None,
    ):
        self.atoms: FancyIndexingList[Atom] = FancyIndexingList()
        if atoms:
            self.add_atoms_from(atoms)
        self.name: str | None = name
        self._lattice_vecs: list[Vec3] | None = None
        self.lattice_vecs = lattice_vecs
        self._cache = LRUCache(maxsize=10)

    @property
    def lattice_vecs(self) -> list[Vec3] | None:
        return self._lattice_vecs

    @lattice_vecs.setter
    def lattice_vecs(self, lattice_vecs: mat_type | None) -> None:
        if lattice_vecs is None:
            self._lattice_vecs = None
            return

        lat_vecs = Mat3(lattice_vecs)
        lat_det = Mat3.det(lat_vecs)
        if lat_det == 0:
            raise ValueError("lattice vectors must not be linearly dependent")
        if lat_det < 0:
            raise ValueError("lattice vectors must form a right-handed basis")

        self._lattice_vecs = lat_vecs

    def __len__(self) -> int:
        return len(self.atoms)

    def __eq__(self, other: Molecule) -> bool:
        return self.atoms == other.atoms

    def __ne__(self, other: Molecule) -> bool:
        return not self == other

    def __contains__(self, atom: Atom) -> bool:
        if not isinstance(atom, Atom):
            raise TypeError("atom must be an Atom object")
        return atom in self.atoms

    def __getitem__(self, index: int | slice | Iterable) -> Atom | Molecule:
        result = self.atoms[index]
        if isinstance(result, Atom):
            return result
        elif isinstance(result, Iterable):
            return Molecule(*result)
        else:
            raise TypeError("Index must be int, slice, list, or tuple")

    def __setitem__(self, index: int, atom: Atom) -> None:
        self.atoms[index] = atom

    def __delitem__(self, idx: int) -> None:
        del self.atoms[idx]

    def __iter__(self):
        for atom in self.atoms:
            if not isinstance(atom, Atom):
                raise TypeError(
                    "Invalid element type: atom must be Atom object\n"
                    + f"{type(atom)=}"
                )
            yield atom

    def __str__(self):
        formula = self.get_formula()
        return f"Molecule({formula})"

    def __repr__(self):
        formula = self.get_formula()
        return f"Molecule({formula})"

    @classmethod
    def sorted(cls, atoms: Molecule | Iterable[Atom], key: callable = None) -> Molecule:
        """
        Create a new molecule by sorting the atoms of the input molecule(s).
        """
        if isinstance(atoms, Molecule):
            atoms = atoms.atoms
        if not all(isinstance(atoms, Atom) for atoms in atoms):
            raise TypeError("All elements must be Molecule objects")
        atoms.sort(key=key)
        mol = cls()
        mol.add_atoms_from(atoms)

        return mol

    def sort(self, key=None) -> None:
        """
        Sort the atoms of the molecule.
        """
        if key is None:
            self.atoms.sort(
                key=lambda atom: (ATOMIC_NUMBERS[atom.symbol], atom.x, atom.y, atom.z)
            )
        else:
            self.atoms.sort(key=key)

    def pop(self, index: int) -> Atom:
        return self.atoms.pop(index)

    def copy(self) -> Molecule:
        return copy.deepcopy(self)

    def is_same_geom(
        self, other: Molecule, rel_tol: float = 1e-5, abs_tol: float = 0.0
    ) -> bool:
        sorted_self = Molecule.sorted(self)
        sorted_other = Molecule.sorted(other)
        return len(self) == len(other) and all(
            atom1.isclose(atom2, rel_tol=rel_tol, abs_tol=abs_tol)
            for atom1, atom2 in zip(
                sorted_self,
                sorted_other,
            )
        )

    @args_to_set
    def filtered_by_symbols(self, symbols: str | Iterable[str]) -> Molecule:
        return Molecule(*[atom for atom in self if atom.symbol in symbols])

    @classmethod
    @args_to_list
    def from_atoms(cls, atoms: Iterable[Atom]) -> Molecule:
        """
        Create a new molecule from Atom objects.
        """
        if not all(isinstance(atom, Atom) for atom in atoms):
            raise TypeError("All elements must be Atom objects")
        return cls(*atoms)

    @classmethod
    def from_file(cls, filepath: str | Path) -> Molecule:
        """
        Create a new molecule from a file.
        auto-detects file format (*.xyz, *.com, *.gjf, *.inp, *.cif, *POSCAR*)
        """
        from molgeom.parsers.parser_selector import read_file

        return read_file(filepath)

    def add_atom(self, atom: Atom) -> None:
        self.atoms.append(atom)

    def add_atoms_from(self, atoms: Iterable[Atom]) -> None:
        """
        Add atoms to the molecule.
        Accepts Iterable containing one or more Atom objects.
        """
        if not isinstance(atoms, Iterable):
            raise TypeError("atoms must be an Iterable of Atom objects")
        not_atoms = [atom for atom in atoms if not isinstance(atom, Atom)]
        if not_atoms:
            raise TypeError(
                "Invalid element type: atoms must be Atom objects " + f"{not_atoms=}"
            )
        self.atoms.extend(atoms)

    def _get_geom_hash(self):
        return hash(tuple((atom.symbol, atom.x, atom.y, atom.z) for atom in self))

    def get_symbols(self) -> list[str]:
        return tuple(atom.symbol for atom in self)

    def get_formula(self) -> str:
        symbol_count = dict()
        for atom in self:
            symbol_count[atom.symbol] = symbol_count.get(atom.symbol, 0) + 1

        formula = "-".join(
            f"{symbol}{count}" if count > 1 else symbol
            for symbol, count in sorted(
                symbol_count.items(), key=lambda x: ATOMIC_NUMBERS[x[0]]
            )
        )
        return formula

    def get_cart_coords(self) -> list[Vec3]:
        return [[atom.x, atom.y, atom.z] for atom in self]

    def set_cart_coords(self, coords: list[Vec3 | list[float | int]]) -> None:
        if len(coords) != len(self):
            raise ValueError(
                "Number of coordinates must match the number of atoms in the molecule"
            )

        for atom, coord in zip(self, coords):
            atom.x, atom.y, atom.z = coord

    def get_frac_coords(self, wrap=False) -> list[Vec3]:
        if self.lattice_vecs is None:
            raise ValueError("Lattice vectors must be set to bound the molecule.")

        return [
            cart2frac(atom.to_Vec3(), self.lattice_vecs, wrap=wrap) for atom in self
        ]

    @cachedmethod(
        lambda self: self._cache,
        key=lambda self, tol=0.15: hashkey(self._get_geom_hash(), tol),
    )
    def get_bonds(
        self,
        tol: float = default_tol,
    ) -> list[dict[tuple[int, int], float]]:

        bonds = list()
        num_atoms = len(self)
        for i in range(num_atoms):
            ai = self[i]
            for j in range(i + 1, num_atoms):
                aj = self[j]
                length = ai.get_bond_length(aj, tol)
                if length is not None:
                    bonds.append({"pair": (i, j), "length": length})

        return bonds

    def get_clusters(self, tol: float = default_tol) -> list[Molecule]:
        G = nx.Graph()
        G.add_nodes_from(range(len(self)))
        G.add_edges_from(bond["pair"] for bond in self.get_bonds(tol))
        copied_mol = self.copy()
        return [copied_mol[list(cluster)] for cluster in nx.connected_components(G)]

    def get_cycles(
        self, length_bound: int | None = None, tol: float = default_tol
    ) -> list[Molecule]:
        G = nx.Graph()
        G.add_edges_from(bond["pair"] for bond in self.get_bonds(tol))
        cycles = nx.simple_cycles(G, length_bound=length_bound)
        return [self[list(cycle)] for cycle in cycles]

    def get_connected_cluster(
        self, atom_idx: int, tol: float = default_tol
    ) -> Molecule:
        G = nx.Graph()
        G.add_edges_from(bond["pair"] for bond in self.get_bonds(tol))
        return self[list(nx.node_connected_component(G, atom_idx))]

    @classmethod
    @args_to_list
    def merged(cls, mols: Molecule | Iterable[Molecule]) -> Molecule:
        """
        Create a new molecule by merging multiple Molecule objects.
        lattice vectors, bonds, and cycles are initialized to None
        """
        if not all(isinstance(mol, Molecule) for mol in mols):
            raise TypeError("All elements must be Molecule objects")
        merged = cls()
        for mol in mols:
            merged.add_atoms_from(mol)

        return merged

    @args_to_list
    def merge(self, mols: Molecule | Iterable[Molecule]) -> Molecule:
        """
        Merge other Molecule objects into this Molecule object.
        resets the lattice vectors, bonds, and cycles
        """
        if not all(isinstance(mol, Molecule) for mol in mols):
            raise TypeError(
                "All elements must be Molecule objects\n"
                + f"{[(i, type(mol)) for i, mol in enumerate(mols) if not isinstance(mol, Molecule)]}"
            )

        for mol in mols:
            self.add_atoms_from(mol)

        self._lattice_vecs = None

    def translate(
        self, trans_vec: Vec3 | list[float | int], with_lattice_vecs: bool = False
    ) -> None:
        if not isinstance(trans_vec, (Vec3, list)):
            raise TypeError("trans_vec must be Vec3 object or list of 3 floats or ints")
        if isinstance(trans_vec, list):
            if len(trans_vec) != 3:
                raise ValueError("trans_vec must be of length 3")
            if not all(isinstance(i, (int, float)) for i in trans_vec):
                raise TypeError("elements of trans_vec must be int or float")
            trans_vec = Vec3(*trans_vec)

        for atom in self:
            atom.translate(trans_vec)

        if with_lattice_vecs and self.lattice_vecs is not None:
            for vec in self.lattice_vecs:
                vec.translate(trans_vec)

    def mirror(self, sx: int, sy: int, sz: int) -> None:
        for atom in self.atoms:
            atom.mirror(sx, sy, sz)

    def mirror_by_plane(self, p1: Vec3, p2: Vec3, p3: Vec3) -> None:
        for atom in self.atoms:
            atom.mirror_by_plane(p1, p2, p3)

    def matmul(self, mat: mat_type, with_lattice_vecs: bool = True) -> None:
        if not is_mat_type(mat):
            raise TypeError("mat must be a 3x3 matrix")

        for atom in self.atoms:
            atom.matmul(mat)

        if with_lattice_vecs and self.lattice_vecs is not None:
            self.lattice_vecs = self.lattice_vecs @ mat

    def rotate_by_axis(
        self,
        axis_point1: vec_type,
        axis_point2: vec_type,
        deg: float,
        with_lattice_vecs: bool = True,
    ) -> None:
        """
        :param axis_point1: One point on the rotation axis
        :param axis_point2: Another point on the rotation axis
        :param angle_degrees: Rotation angle (degrees)
        """

        for atom in self.atoms:
            atom.rotate_by_axis(axis_point1, axis_point2, deg)

        if with_lattice_vecs and self.lattice_vecs is not None:
            for vec in self.lattice_vecs:
                vec.rotate_by_axis(axis_point1, axis_point2, deg)

    def replicate(self, rep_a: list[int], rep_b: list[int], rep_c: list[int]) -> None:
        """
        Replicate the molecule in the a, b, and c directions.
        """
        if self.lattice_vecs is None:
            raise ValueError("Lattice vectors must be set to replicate the molecule.")

        if not all(isinstance(rep, list) for rep in (rep_a, rep_b, rep_c)):
            raise TypeError(
                "Replication vectors must be of length 2 list\n"
                + f"{(rep_a, rep_b, rep_c)=}"
            )

        if len(rep_a) != 2 or len(rep_b) != 2 or len(rep_c) != 2:
            raise ValueError(
                "Replication vectors must be of length 2 list\n"
                + f"{(rep_a, rep_b, rep_c)=}"
            )

        if rep_a[0] >= rep_a[1] or rep_b[0] >= rep_b[1] or rep_c[0] >= rep_c[1]:
            raise ValueError(
                "Replication vectors must be of the form [start, end], start < end\n"
                + f"{(rep_a, rep_b, rep_c)=}"
            )

        tmp_mol = self.copy()
        self.atoms = FancyIndexingList()
        for i in range(rep_a[0], rep_a[1]):
            for j in range(rep_b[0], rep_b[1]):
                for k in range(rep_c[0], rep_c[1]):

                    trans_vec = (
                        i * tmp_mol.lattice_vecs[0]
                        + j * tmp_mol.lattice_vecs[1]
                        + k * tmp_mol.lattice_vecs[2]
                    )
                    mol_copied = tmp_mol.copy()
                    mol_copied.translate(trans_vec)
                    self.merge(mol_copied)

        self.lattice_vecs = tmp_mol.lattice_vecs

    def is_inside_cell(self, atom: Atom) -> bool:
        """
        Check if the atom is inside the parallelepiped defined by the lattice vectors.

        let v = u1*a1 + u2*a2 + u3*a3, where a1, a2, a3 are lattice vectors
        mat = [a2×a3, a3×a1, a1×a2]
        to check if v is inside the cell, we need to check if 0 <= u1, u2, u3 < 1
        U = [u1, u2, u3] = 1/V * mat * v
        """
        if not isinstance(atom, Atom):
            raise TypeError(f"Expected Atom, got {type(atom).__name__}")

        U = atom.get_frac_coords(self.lattice_vecs)
        return all(0 <= u < 1 for u in U)

    def remove_duplicates(self, tol: float = default_tol) -> None:
        """
        Remove duplicate atoms by combining close atoms with the same symbol
        into their centroid.
        Args:
            tol (float): Tolerance for distance to consider atoms as duplicates.
        """
        unique_atoms = FancyIndexingList()

        symbols = set(a.symbol for a in self.atoms)
        for symbol in symbols:
            indices = [i for i, atom in enumerate(self) if atom.symbol == symbol]
            sub_atoms = self[indices]

            processed = set()
            for i, atom in enumerate(sub_atoms):
                if i in processed:
                    continue

                cluster = [i]
                for j, other_atom in enumerate(sub_atoms):
                    if j != i and atom.distance_to(other_atom) < tol:
                        cluster.append(j)

                processed.update(cluster)

                cluster_atoms = sub_atoms[cluster]
                centroid = cluster_atoms.center_of_mass()

                unique_atoms.append(Atom(symbol, *centroid))

        self.atoms = unique_atoms

    def wrap_to_cell(self) -> None:
        """
        Wrap the molecule to the cell defined by the lattice vectors.
        """
        if self.lattice_vecs is None:
            raise ValueError("Lattice vectors must be set to bound the molecule.")

        frac_coords = self.get_frac_coords(wrap=True)
        for atom, frac_coord in zip(self, frac_coords):
            cart_coords = frac2cart(frac_coord, self.lattice_vecs)
            atom.x, atom.y, atom.z = cart_coords

    def replicated_from_xyz_str(self, xyz_str: str, wrap: bool = True) -> Molecule:
        """
        Replicate the molecule from an xyz string in cif format.
        args:
            xyz_str (str): xyz string of symmetry operation. e.g. 'x, y, z', '-y, -x+3/4, z+1/2',
            wrap (bool): wrap the molecule to the cell after replication
        """

        if self.lattice_vecs is None:
            raise ValueError("Lattice vectors must be set to replicate the molecule.")

        rot_mat, trans_vec_frac = symmop_from_xyz_str(xyz_str)

        new_mol = self.copy()
        new_mol.matmul(rot_mat, with_lattice_vecs=False)
        new_mol.translate(frac2cart(trans_vec_frac, self.lattice_vecs))

        if wrap:
            new_mol.wrap_to_cell()

        return new_mol

    def total_mass(self) -> float:
        return sum(atom.mass for atom in self)

    def center_of_mass(self) -> Vec3:
        total_mass = self.total_mass()
        com_x = sum(atom.x * atom.mass for atom in self) / total_mass
        com_y = sum(atom.y * atom.mass for atom in self) / total_mass
        com_z = sum(atom.z * atom.mass for atom in self) / total_mass
        return Vec3(com_x, com_y, com_z)

    def nuclear_repulsion(self) -> float:
        """
        Calculate the nuclear repulsion energy of the molecule.
        uses gaussian16 constants
        """
        nuclrep = 0.0
        for i in range(len(self.atoms)):
            for j in range(i + 1, len(self.atoms)):
                dist_angst = self.atoms[i].distance_to(self.atoms[j])
                dist_bohr = dist_angst * ANGST2BOHR_GAU16
                nuclrep += (
                    self.atoms[i].atomic_number
                    * self.atoms[j].atomic_number
                    / dist_bohr
                )
        return nuclrep

    def nuclrep(self) -> float:
        """
        Alias for nuclear_repulsion method
        """
        return self.nuclear_repulsion()

    def electrostatic_energy(self) -> float:
        """
        Calculate the electrostatic energy of the molecule.
        """
        elec_energy = 0.0
        for i in range(len(self.atoms)):
            for j in range(i + 1, len(self.atoms)):
                dist_angst = self.atoms[i].distance_to(self.atoms[j])
                dist_bohr = dist_angst * ANGST2BOHR_GAU16
                if self.atoms[i].charge is None or self.atoms[j].charge is None:
                    raise ValueError("All atoms must have their charge set.")
                elec_energy += self.atoms[i].charge * self.atoms[j].charge / dist_bohr
        return elec_energy

    def to_xyz(self) -> str:
        return "\n".join([atom.to_xyz() for atom in self])

    def to_dict(self) -> dict:
        mol_dict = dict()
        for i, atom in enumerate(self):
            mol_dict[i] = atom.to_dict()
        return mol_dict

    def write_to_xyz(self, filepath: str) -> None:
        filepath = os.path.expanduser(filepath.strip())
        with open(filepath, "w") as f:
            f.write(f"{len(self)}\n")
            f.write(self.get_formula() + "\n")
            f.write(self.to_xyz())
        print(f"File written to {filepath}")

    @staticmethod
    def write_to_xyzs(
        mols: list[Molecule],
        filepath: str | Path,
    ) -> None:
        filepath = Path(str(filepath).strip()).expanduser()
        with open(filepath, "w") as f:
            for mol in mols:
                f.write(f"{len(mol)}\n")
                f.write(mol.get_formula() + "\n")
                f.write(mol.to_xyz() + "\n")
        print(f"File written to {filepath}")

    def write_to_gaussian_input(
        self, filepath: str | Path, head: str | None = None, tail: str | None = None
    ) -> None:
        filepath = Path(str(filepath).strip()).expanduser()
        with open(filepath, "w") as f:
            if head is not None:
                f.write(head)
                f.write("\n")
            else:
                f.write("#p B3LYP\n")
                f.write("\n")
                f.write(f"{self.get_formula()}\n")
                f.write("\n")
                f.write("0 1\n")
            f.write(self.to_xyz() + "\n")
            if self.lattice_vecs is not None:
                for vec in self.lattice_vecs:
                    f.write(
                        f"{'Tv':2s} {vec[0]:19.12f} {vec[1]:19.12f} {vec[2]:19.12f}\n"
                    )
            if tail is not None:
                f.write(tail)
                f.write("\n")
        print(f"File written to {filepath}")

    def write_to_gamess_input(
        self, filepath: str | Path, head: str | None = None, tail: str | None = None
    ) -> None:
        filepath = Path(str(filepath).strip()).expanduser()
        with open(filepath, "w") as f:
            if head is not None:
                f.write(head)
            else:
                f.write("$DATA\n")
                f.write(f"{self.get_formula()}\n")
                f.write("C1\n")
            for i in range(len(self)):
                f.write(
                    f" {self[i].symbol:2s}  {float(self[i].atomic_number):.1f}  {self[i].x:19.12f} {self[i].y:19.12f} {self[i].z:19.12f}\n"
                )
            if tail is not None:
                f.write(tail)
            else:
                f.write("$END\n")
        print(f"File written to {filepath}")

    def write_to_poscar(
        self, filepath: str | Path, frac: bool = True, wrap: bool = False
    ) -> None:
        filepath = Path(str(filepath).strip()).expanduser()
        with open(filepath, "w") as f:
            f.write(f"{self.get_formula()}\n")
            f.write("1.0\n")
            f.write(
                f"{self.lattice_vecs[0][0]:19.12f} {self.lattice_vecs[0][1]:19.12f} {self.lattice_vecs[0][2]:19.12f}\n"
            )
            f.write(
                f"{self.lattice_vecs[1][0]:19.12f} {self.lattice_vecs[1][1]:19.12f} {self.lattice_vecs[1][2]:19.12f}\n"
            )
            f.write(
                f"{self.lattice_vecs[2][0]:19.12f} {self.lattice_vecs[2][1]:19.12f} {self.lattice_vecs[2][2]:19.12f}\n"
            )
            # remove duplicates and keep the order
            unique_symbols = list(dict.fromkeys([atom.symbol for atom in self]))
            f.write(" ".join(unique_symbols) + "\n")
            symbol_count = dict()
            for symbol in unique_symbols:
                symbol_count[symbol] = sum(atom.symbol == symbol for atom in self)
            f.write(
                " ".join(str(symbol_count[symbol]) for symbol in unique_symbols) + "\n"
            )
            if frac:
                f.write("Direct\n")
                coords = self.get_frac_coords(wrap=wrap)
            else:
                f.write("Cartesian\n")
                coords = self.coords
            for coord in coords:
                f.write(f"{coord[0]:19.12f} {coord[1]:19.12f} {coord[2]:19.12f}\n")
        print(f"File written to {filepath}")

    def write_to_cif(self, filepath: str | Path, wrap: bool = False) -> None:
        filepath = Path(str(filepath).strip()).expanduser()
        with open(filepath, "w") as f:
            f.write("data_molecule\n")
            f.write("########################\n")
            f.write("# generated by molgeom #\n")
            f.write("########################\n")
            f.write("\n")
            f.write("loop_\n")
            f.write("_space_group_symop_operation_xyz\n")
            f.write(" 'x, y, z'\n")
            f.write("\n")
            a, b, c, alpha, beta, gamma = lat_vecs_to_lat_params(self.lattice_vecs)
            f.write(f"_cell_length_a {a:.9f}\n")
            f.write(f"_cell_length_b {b:.9f}\n")
            f.write(f"_cell_length_c {c:.9f}\n")
            f.write(f"_cell_angle_alpha {alpha:.9f}\n")
            f.write(f"_cell_angle_beta {beta:.9f}\n")
            f.write(f"_cell_angle_gamma {gamma:.9f}\n")
            f.write("\n")
            f.write("loop_\n")
            f.write("_atom_site_label\n")
            f.write("_atom_site_type_symbol\n")
            f.write("_atom_site_fract_x\n")
            f.write("_atom_site_fract_y\n")
            f.write("_atom_site_fract_z\n")
            frac_coords = self.get_frac_coords(wrap=wrap)
            for i in range(len(self)):
                f.write(
                    f"{i+1:3d} {self[i].symbol:2s} {frac_coords[i][0]:19.12f} {frac_coords[i][1]:19.12f} {frac_coords[i][2]:19.12f}\n"
                )
        print(f"File written to {filepath}")

    def write_to_mol(self, filepath: str | Path) -> None:
        """Write molecule to MOL V2000 format file.

        Creates a MOL file following V2000 format specification.
        Includes atomic coordinates, bonds, and charges.
        """
        bonds = self.get_bonds()

        filepath = Path(str(filepath).strip()).expanduser()
        with open(filepath, "w") as f:
            # Header block
            f.write(f"{self.get_formula()}\n")
            f.write("  Generated by molgeom\n")
            f.write("\n")  # Comment line

            # Counts line with actual bond count
            f.write(f"{len(self):3d}{len(bonds):3d}  0  0  0  0  0  0  0  0999 V2000\n")

            # Atom block
            for atom in self:
                charge = atom.charge if atom.charge is not None else 0
                f.write(
                    f"{atom.x:10.4f}{atom.y:10.4f}{atom.z:10.4f} {atom.symbol:<3s}"
                    f" 0  0  0  0  0  0  0  0  0{charge:3d}  0  0  0\n"
                )

            # Bond block - assuming single bonds for now
            for bond in bonds:
                i, j = bond["pair"]
                # MOL format uses 1-based indexing
                f.write(f"{i+1:3d}{j+1:3d}  1  0  0  0  0\n")

            # End marker
            f.write("M  END\n")

        print(f"File written to {filepath}")

    @staticmethod
    def write_to_sdf(
        mols: list[Molecule],
        filepath: str | Path,
        properties: dict[str, dict[str, str]] | None = None,
    ) -> None:
        """Write multiple Molecule objects to SDF format file.

        Args:
            mols: list of Molecule objects to write
            filepath: Output file path
            properties: Optional dictionary of molecular properties
                        Format: {mol_index: {property_name: property_value}}
        """

        filepath = Path(str(filepath).strip()).expanduser()
        with open(filepath, "w") as f:
            for i, mol in enumerate(mols):
                bonds = mol.get_bonds()

                # Write MOL format content
                f.write(f"{mol.get_formula()}\n")
                f.write("  Generated by molgeom\n")
                f.write("\n")
                f.write(
                    f"{len(mol):3d}{len(bonds):3d}  0  0  0  0  0  0  0  0999 V2000\n"
                )

                # Atom block
                for atom in mol:
                    charge = atom.charge if atom.charge is not None else 0
                    f.write(
                        f"{atom.x:10.4f}{atom.y:10.4f}{atom.z:10.4f} {atom.symbol:<3s}"
                        f" 0  0  0  0  0  0  0  0  0{charge:3d}  0  0  0\n"
                    )

                # Bond block
                for bond in bonds:
                    i, j = bond["pair"]
                    digits = len(str(len(mol)))
                    f.write(f"{i+1:>{digits+2}d}{j+1:>{digits+2}d}  1  0  0  0  0\n")

                f.write("M  END\n")

                # Write properties if provided
                if properties and i in properties:
                    for name, value in properties[i].items():
                        f.write(f"> <{name}>\n")
                        f.write(f"{value}\n")
                        f.write("\n")

                # Add molecule separator
                f.write("$$$$\n")

        print(f"File written to {filepath}")

    def view_mol(self, cleanup: bool = True, prefer_notebook: bool = True) -> None:
        """
        View molecular geometry using 3Dmol.js in a browser.
        args:
            cleanup: bool
                If True, removes temporary files after viewing (default: True)
            prefer_notebook: bool
                If True, opens in Jupyter notebook if available (default: True)
        """
        xyz_data = f"{len(self)}\n{self.name or str(self)}\n{self.to_xyz()}"
        view_mol(
            xyz_mol_data=xyz_data, cleanup=cleanup, prefer_notebook=prefer_notebook
        )

    @staticmethod
    def view_mols(
        mols: list[Molecule], cleanup: bool = True, prefer_notebook: bool = True
    ) -> None:
        """
        View multiple molecular geometries using 3Dmol.js in a browser.
        args:
            mols: list[Molecule]
                List of Molecule objects to view
            cleanup: bool
                If True, removes temporary files after viewing (default: True)
            prefer_notebook: bool
                If True, opens in Jupyter notebook if available (default: True)
        """
        xyz_data = "\n".join(
            [f"{len(mol)}\n{mol.name or str(mol)}\n{mol.to_xyz()}" for mol in mols]
        )
        view_mol(
            xyz_mol_data=xyz_data, cleanup=cleanup, prefer_notebook=prefer_notebook
        )



import re
from pathlib import Path

from molgeom.atom import Atom
from molgeom.molecule import Molecule
from molgeom.parsers.parser_tools import (
    symbol_xyz_regex,
    is_valid_xyz_line,
    remove_trailing_empty_lines,
    validate_filepath,
    zopen,
)


def from_xyz_str(content: str) -> Molecule:
    mole = Molecule()

    lines = remove_trailing_empty_lines(content.strip().split("\n"))

    # replace tabs, non-breaking spaces, and multiple spaces with single space
    for i in range(len(lines)):
        lines[i] = re.sub(r"[\s\t\xa0]+", " ", lines[i])

    first_line = lines[0].strip()
    mol_struc_lines = []
    if re.fullmatch(r"\d+", first_line):
        num_atoms = int(first_line)
        _comment_line = lines[1].strip()
        mol_struc_lines = lines[2:]
    elif re.fullmatch(symbol_xyz_regex, first_line):
        mol_struc_lines = lines
        num_atoms = len(mol_struc_lines)
    else:
        raise ValueError(f"Invalid file format in first line: \n{first_line}")

    for line in mol_struc_lines:
        if not is_valid_xyz_line(line):
            raise ValueError(f"Invalid line format: \n{line}")

        data = line.strip().split()
        atom = Atom(
            symbol=data[0], x=float(data[1]), y=float(data[2]), z=float(data[3])
        )
        mole.atoms.append(atom)

    if num_atoms != len(mole):
        raise ValueError(
            f"Number of atoms specified ({num_atoms}) "
            + f"does not match number of atoms read ({len(mole)})."
        )

    return mole


def xyz_parser(filepath: str | Path) -> Molecule:
    filepath = validate_filepath(filepath)
    with zopen(filepath, mode="rt", encoding="utf-8") as file:
        content = file.read()
    mol = from_xyz_str(content)
    mol.name = filepath.stem
    return mol



from __future__ import annotations

import sys
from pathlib import Path

from molgeom.molecule import Molecule
from molgeom.parsers.parser_tools import validate_filepath
from molgeom.parsers.cif import cif_parser
from molgeom.parsers.gamess import gms_inp_parser
from molgeom.parsers.gaussian import gau_inp_parser
from molgeom.parsers.poscar import poscar_parser
from molgeom.parsers.xyz import xyz_parser
from molgeom.parsers.mol import mol_parser
from molgeom.parsers.sdf import sdf_parser


def read_file(filepath: str | Path) -> Molecule:
    """Read file and return Molecule object."""
    filepath = validate_filepath(filepath)

    ext_parser_map = {
        ".xyz": xyz_parser,
        ".com": gau_inp_parser,
        ".gjf": gau_inp_parser,
        ".inp": gms_inp_parser,
        ".cif": cif_parser,
        ".mol": mol_parser,
        ".sdf": sdf_parser,
    }

    for ext in ext_parser_map:
        if ext in filepath.suffixes:
            return ext_parser_map[ext](filepath)

    if "poscar" in str(filepath).lower():
        return poscar_parser(filepath)

    raise RuntimeError(
        f'file extension for "{filepath.name}" '
        + "is not supported or extensionless file"
    )


def main():
    if len(sys.argv) < 2:
        print("Usage: python script.py <xyz_file_path>")
        sys.exit(1)

    print()
    filepaths = sys.argv[1:]
    for filepath in filepaths:
        print(filepath)
        mole = read_file(filepath)
        print(mole.to_xyz())


if __name__ == "__main__":
    main()



import math
from .mat3 import Mat3
from .vec3 import Vec3


def lat_params_to_lat_vecs(a, b, c, alpha, beta, gamma, angle_in_degrees=True) -> Mat3:
    if angle_in_degrees:
        alpha = math.radians(alpha)
        beta = math.radians(beta)
        gamma = math.radians(gamma)

    cosa = math.cos(alpha)
    cosb = math.cos(beta)
    cosg = math.cos(gamma)
    sing = math.sin(gamma)
    volume = math.sqrt(
        1.0 - cosa**2.0 - cosb**2.0 - cosg**2.0 + 2.0 * cosa * cosb * cosg
    )

    mat = Mat3(
        [
            a * Vec3(1, 0, 0),
            b * Vec3(cosg, sing, 0),
            c * Vec3(cosb, (cosa - cosb * cosg) / sing, volume / sing),
        ]
    )

    return mat


def lat_vecs_to_lat_params(lattice_vecs: Mat3) -> tuple:
    a = lattice_vecs[0].norm()
    b = lattice_vecs[1].norm()
    c = lattice_vecs[2].norm()
    alpha = math.degrees(lattice_vecs[1].angle(lattice_vecs[2]))
    beta = math.degrees(lattice_vecs[0].angle(lattice_vecs[2]))
    gamma = math.degrees(lattice_vecs[0].angle(lattice_vecs[1]))

    return a, b, c, alpha, beta, gamma


def wrap_frac_coords(frac_coords: Vec3) -> Vec3:
    return frac_coords % 1.0


def cart2frac(cart_coords: Vec3, lattice_vecs: Mat3, wrap=False) -> Vec3:
    frac_coords = lattice_vecs.inv().T() @ cart_coords
    if wrap:
        frac_coords = wrap_frac_coords(frac_coords)
    return frac_coords


def frac2cart(frac_coords: Vec3, lattice_vecs: Mat3) -> Vec3:
    return lattice_vecs.T() @ frac_coords



from __future__ import annotations
from copy import deepcopy

import math
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .mat3 import Mat3

# import numpy as np
# from scipy.spatial.transform import Rotation as R


class Vec3:
    def __init__(self, x: int | float, y: int | float, z: int | float) -> None:
        if not all(isinstance(i, (int, float)) for i in (x, y, z)):
            raise TypeError("x, y, z must be an int or float")
        self._x = float(x)
        self._y = float(y)
        self._z = float(z)

    @property
    def x(self) -> float:
        return self._x

    @x.setter
    def x(self, value: int | float) -> None:
        if not isinstance(value, (int, float)):
            raise TypeError("x must be an int or float")
        self._x = float(value)

    @property
    def y(self) -> float:
        return self._y

    @y.setter
    def y(self, value: int | float) -> None:
        if not isinstance(value, (int, float)):
            raise TypeError("y must be an int or float")
        self._y = float(value)

    @property
    def z(self) -> float:
        return self._z

    @z.setter
    def z(self, value: int | float) -> None:
        if not isinstance(value, (int, float)):
            raise TypeError("z must be an int or float")
        self._z = float(value)

    @classmethod
    def from_list(cls, vec: list[int | float]) -> Vec3:
        if not is_vec_type(vec):
            raise ValueError("vec must be a list of 3 int or float")
        return cls(*vec)

    def to_list(self) -> list[float]:
        return [self.x, self.y, self.z]

    def to_dict(self) -> dict[str, float]:
        return {"x": self.x, "y": self.y, "z": self.z}

    def __str__(self) -> str:
        return f"{self.x:19.12f} {self.y:19.12f} {self.z:19.12f}"

    def __repr__(self) -> str:
        return f"Vec3({self.x:.12f}, {self.y:.12f}, {self.z:.12f})"

    def __getitem__(self, index: int) -> float:
        return (self.x, self.y, self.z)[index]

    def __setitem__(self, index: int, value: int | float) -> None:
        if not isinstance(value, (int, float)):
            raise TypeError("value must be an int or float")
        if index in (0, -3):
            self.x = value
        elif index in (1, -2):
            self.y = value
        elif index in (2, -1):
            self.z = value
        else:
            raise IndexError("index out of range")

    def __eq__(self, other: Vec3) -> bool:
        return self.x == other.x and self.y == other.y and self.z == other.z

    def __ne__(self, other: Vec3) -> bool:
        return not self == other

    def __neg__(self) -> Vec3:
        return Vec3(-self.x, -self.y, -self.z)

    def __add__(self, other: Vec3) -> Vec3:
        other_class_check_for_operand(self, other, "+")
        copied_self = self.copy()
        copied_self.x += other.x
        copied_self.y += other.y
        copied_self.z += other.z
        return copied_self

    def __radd__(self, other: Vec3) -> Vec3:
        return other.__add__(self)

    def __iadd__(self, other: Vec3) -> Vec3:
        self.x += other.x
        self.y += other.y
        self.z += other.z
        return self

    def __sub__(self, other: Vec3) -> Vec3:
        other_class_check_for_operand(self, other, "-")
        copied_self = self.copy()
        return copied_self.__isub__(other)

    def __rsub__(self, other: Vec3) -> Vec3:
        copied_self = self.copy()
        copied_self.x = other.x - copied_self.x
        copied_self.y = other.y - copied_self.y
        copied_self.z = other.z - copied_self.z
        return copied_self

    def __isub__(self, other: Vec3) -> Vec3:
        self.x -= other.x
        self.y -= other.y
        self.z -= other.z
        return self

    def __mul__(self, value: int | float | Vec3) -> Vec3:
        if not isinstance(value, (int, float, Vec3)):
            raise TypeError("value must be an int, float or Vec3")
        if isinstance(value, (int, float)):
            return Vec3(self.x * value, self.y * value, self.z * value)

        copied_self = self.copy()
        copied_self.x *= value.x
        copied_self.y *= value.y
        copied_self.z *= value.z
        return copied_self

    def __rmul__(self, value: int | float | Vec3) -> Vec3:
        return self.__mul__(value)

    def __imul__(self, value: int | float | Vec3) -> Vec3:
        if isinstance(value, (int, float)):
            self.x *= value
            self.y *= value
            self.z *= value
            return self
        if isinstance(value, Vec3):
            self.x *= value.x
            self.y *= value.y
            self.z *= value.z
            return self
        raise TypeError("value must be an int, float or Vec3")

    def __truediv__(self, value: int | float | Vec3) -> Vec3:
        if not isinstance(value, (int, float, Vec3)):
            raise TypeError("value must be an int, float or Vec3")
        if isinstance(value, (int, float)):
            return Vec3(self.x / value, self.y / value, self.z / value)

        copied_self = self.copy()
        copied_self.x /= value.x
        copied_self.y /= value.y
        copied_self.z /= value.z

        return copied_self

    def __itruediv__(self, value: int | float | Vec3) -> Vec3:
        if isinstance(value, (int, float)):
            self.x /= value
            self.y /= value
            self.z /= value
            return self
        if isinstance(value, Vec3):
            self.x /= value.x
            self.y /= value.y
            self.z /= value.z
            return self
        raise TypeError("value must be an int, float or Vec3")

    def __floordiv__(self, value: int | float | Vec3) -> Vec3:
        if not isinstance(value, (int, float, Vec3)):
            raise TypeError("value must be an int, float or Vec3")
        if isinstance(value, (int, float)):
            return Vec3(self.x // value, self.y // value, self.z // value)

        copied_self = self.copy()
        copied_self.x //= value.x
        copied_self.y //= value.y
        copied_self.z //= value.z
        return copied_self

    def __ifloordiv__(self, value: int | float | Vec3) -> Vec3:
        if isinstance(value, (int, float)):
            self.x //= value
            self.y //= value
            self.z //= value
            return self
        if isinstance(value, Vec3):
            self.x //= value.x
            self.y //= value.y
            self.z //= value.z
            return self
        raise TypeError("value must be an int, float or Vec3")

    def __mod__(self, value: int | float | Vec3) -> Vec3:
        if not isinstance(value, (int, float, Vec3)):
            raise TypeError("value must be an int, float or Vec3")
        if isinstance(value, (int, float)):
            return Vec3(self.x % value, self.y % value, self.z % value)

        copied_self = self.copy()
        copied_self.x %= value.x
        copied_self.y %= value.y
        copied_self.z %= value.z
        return copied_self

    def __imod__(self, value: int | float | Vec3) -> Vec3:
        if isinstance(value, (int, float)):
            self.x %= value
            self.y %= value
            self.z %= value
            return self
        if isinstance(value, Vec3):
            self.x %= value.x
            self.y %= value.y
            self.z %= value.z
            return self
        raise TypeError("value must be an int, float or Vec3")

    def __len__(self) -> int:
        return 3

    def copy(self) -> Vec3:
        return deepcopy(self)

    def dot(self, other: vec_type) -> float:
        other = vectorize_arg(other)
        return self.x * other.x + self.y * other.y + self.z * other.z

    def cross(self, other: vec_type) -> Vec3:
        return Vec3(
            self.y * other.z - self.z * other.y,
            self.z * other.x - self.x * other.z,
            self.x * other.y - self.y * other.x,
        )

    def distance_to(self, other: Vec3) -> float:
        other = vectorize_arg(other)
        return math.sqrt(
            (self.x - other.x) ** 2 + (self.y - other.y) ** 2 + (self.z - other.z) ** 2
        )

    def isclose(self, other: Vec3, rel_tol: float = 1e-5, abs_tol: float = 0.0) -> bool:
        other = vectorize_arg(other)
        self_coords = [self.x, self.y, self.z]
        other_coords = [other.x, other.y, other.z]
        return all(
            math.isclose(
                self_coords[i],
                other_coords[i],
                rel_tol=rel_tol,
                abs_tol=abs_tol,
            )
            for i in range(3)
        )

    def norm(self) -> float:
        return math.sqrt(self.x**2 + self.y**2 + self.z**2)

    def normsq(self) -> float:
        return self.x**2 + self.y**2 + self.z**2

    def normalize(self) -> None:
        self /= self.norm()

    def angle(self, other: Vec3) -> float:
        if not isinstance(other, Vec3):
            raise TypeError(f"Expected Vec3, got {type(other)}")
        other = vectorize_arg(other)
        dot = self.dot(other)
        cos_theta = dot / (self.norm() * other.norm())
        return math.acos(cos_theta)

    def mid_point(self, other: Vec3) -> Vec3:
        return (self + other) / 2

    def translate(self, trans_vec) -> None:
        trans_vec = vectorize_arg(trans_vec)
        self.x += trans_vec.x
        self.y += trans_vec.y
        self.z += trans_vec.z

    def mirror(self, sx: int, sy: int, sz: int) -> None:
        if sx not in (-1, 1) or sy not in (-1, 1) or sz not in (-1, 1):
            raise ValueError("mirror factors must be 1 or -1")
        self.x *= sx
        self.y *= sy
        self.z *= sz

    def mirror_by_plane(self, p1: vec_type, p2: vec_type, p3: vec_type) -> None:
        if not all(is_vec_type(p) for p in (p1, p2, p3)):
            raise TypeError("p1, p2, p3 must be Vec3 instances")

        p1 = vectorize_arg(p1)
        p2 = vectorize_arg(p2)
        p3 = vectorize_arg(p3)

        v1 = p2 - p1
        v2 = p3 - p1

        normal = v1.cross(v2)

        if normal.norm() < 1e-6:
            raise ValueError("the three points are collinear")

        # calculate the projection of the point on the mirror plane
        projection = self - normal * (self - p1).dot(normal) / normal.normsq()

        # calculate the mirror image of the point
        self += 2 * (projection - self)

    def matmul(self, mat: mat_type) -> None:
        x_rot = mat[0][0] * self.x + mat[0][1] * self.y + mat[0][2] * self.z
        y_rot = mat[1][0] * self.x + mat[1][1] * self.y + mat[1][2] * self.z
        z_rot = mat[2][0] * self.x + mat[2][1] * self.y + mat[2][2] * self.z

        self.x = x_rot
        self.y = y_rot
        self.z = z_rot

    def rotate_by_axis(
        self,
        axis_point1: vec_type,
        axis_point2: vec_type,
        deg: float | int,
    ) -> None:
        """
        :param axis_point1: One point on the rotation axis
        :param axis_point2: Another point on the rotation axis
        :param deg: Rotation angle (degrees)
        """

        # if scipy and numpy is available, you can use the following code
        """
        angle_radians = np.deg2rad(deg)
        axis_vector = np.array(
            [
                axis_point2.x - axis_point1.x,
                axis_point2.y - axis_point1.y,
                axis_point2.z - axis_point1.z,
            ]
        )
        axis_unit_vector = axis_vector / np.linalg.norm(axis_vector)
        point = np.array([*self]) - np.array([*axis_point1])
        rotation = R.from_rotvec(angle_radians * axis_unit_vector)
        rotated_point = rotation.apply(point)
        rotated_point += np.array([*axis_point1])
        self.x, self.y, self.z = rotated_point.tolist()
        """

        axis_point1 = vectorize_arg(axis_point1)
        axis_point2 = vectorize_arg(axis_point2)

        angle_radians = math.radians(deg)
        axis_vector = axis_point2 - axis_point1
        axis_unit_vector = axis_vector / axis_vector.norm()

        # translate to move the rotation axis to the origin
        self -= axis_point1

        # calculate elements of rotation matrix
        cos_theta = math.cos(angle_radians)
        sin_theta = math.sin(angle_radians)
        one_minus_cos = 1 - cos_theta

        # Rodrigues' rotation formula
        x_rot = (
            (cos_theta + axis_unit_vector.x**2 * one_minus_cos) * self.x
            + (
                axis_unit_vector.x * axis_unit_vector.y * one_minus_cos
                - axis_unit_vector.z * sin_theta
            )
            * self.y
            + (
                axis_unit_vector.x * axis_unit_vector.z * one_minus_cos
                + axis_unit_vector.y * sin_theta
            )
            * self.z
        )

        y_rot = (
            (
                axis_unit_vector.y * axis_unit_vector.x * one_minus_cos
                + axis_unit_vector.z * sin_theta
            )
            * self.x
            + (cos_theta + axis_unit_vector.y**2 * one_minus_cos) * self.y
            + (
                axis_unit_vector.y * axis_unit_vector.z * one_minus_cos
                - axis_unit_vector.x * sin_theta
            )
            * self.z
        )

        z_rot = (
            (
                axis_unit_vector.z * axis_unit_vector.x * one_minus_cos
                - axis_unit_vector.y * sin_theta
            )
            * self.x
            + (
                axis_unit_vector.z * axis_unit_vector.y * one_minus_cos
                + axis_unit_vector.x * sin_theta
            )
            * self.y
            + (cos_theta + axis_unit_vector.z**2 * one_minus_cos) * self.z
        )

        # translate back to the original position
        self.x = x_rot + axis_point1.x
        self.y = y_rot + axis_point1.y
        self.z = z_rot + axis_point1.z


# can be used as a type hint. not isinstance() check
vec_type = Vec3 | list[int | float]
mat_type = list[Vec3 | list[int | float]]


def vectorize_arg(arg: vec_type) -> Vec3:
    if isinstance(arg, Vec3):
        return arg
    if isinstance(arg, list) and len(arg) == 3:
        if all(isinstance(i, (int, float)) for i in arg):
            return Vec3(*arg)
    raise ValueError("arg must be a Vec3 instance or a list of 3 int or float")


def other_class_check_for_operand(self, other, operand: str) -> None:
    if not isinstance(other, self.__class__):
        return NotImplementedError(
            f"unsupported operand type(s) for {operand}:"
            + f"'{self.__class__.__name__}' and '{other.__class__.__name__}'"
        )


def is_vec_type(vec: list[float]) -> bool:
    return (
        isinstance(vec, list)
        and len(vec) == 3
        and all(isinstance(i, (int, float)) for i in [*vec])
    )


def is_mat_type(mat: list[list[float]] | Mat3) -> bool:
    return isinstance(mat, Mat3) or (
        isinstance(mat, list)
        and len(mat) == 3
        and all(is_vec_type(vec) for vec in mat if isinstance(vec, list))
    )



import re
from pathlib import Path
from collections import deque

from molgeom.atom import Atom
from molgeom.molecule import Molecule
from molgeom.parsers.parser_tools import (
    is_valid_gms_xyz_line,
    remove_trailing_empty_lines,
    validate_filepath,
    zopen,
)


def from_gms_inp_str(content: str) -> Molecule:
    mol = Molecule()

    lines = deque(remove_trailing_empty_lines(content.strip().split("\n")))

    # replace tabs, non-breaking spaces, and multi spaces with single space
    for i in range(len(lines)):
        lines[i] = re.sub(r"[\s\t\xa0]+", " ", lines[i])

    # input description section
    input_descriptions = []
    while lines and not lines[0].strip().upper().startswith("$DATA"):
        if lines[0].strip().startswith("!"):
            lines.popleft()
            continue
        input_descriptions.append(lines.popleft().strip())

    # $DATA group
    lines.popleft()
    if not lines[0].strip() or not lines[1].strip():
        raise ValueError(
            "inp_parser DATA group : invalid file format \n"
            + f"{lines[0]}\n"
            + f"{lines[1]}"
        )
    _title = lines.popleft().strip()
    _group_naxis = lines.popleft().strip()

    # atom cartesian coordinates
    while lines and lines[0].strip() != "$END":
        if not is_valid_gms_xyz_line(lines[0]):
            raise ValueError(
                "inp_parser atom cartesian coords :"
                + f"invalid file format \n{lines[0]}"
            )
        data = lines.popleft().strip().split()
        atom = Atom(
            symbol=data[0], x=float(data[2]), y=float(data[3]), z=float(data[4])
        )
        mol.add_atom(atom)

    return mol


def extract_head_tail_from_gms_inp(filepath: str | Path) -> tuple[str, str]:
    file_head = []
    file_tail = []
    filepath = Path(filepath)
    if not filepath.exists():
        raise FileNotFoundError(f"{filepath} do not exist")
    if not filepath.is_file():
        raise ValueError(f"{filepath} is not a file")
    with zopen(filepath, mode="rt", encoding="utf-8") as file:
        while True:
            line = file.readline()
            if line.strip().upper().startswith("$DATA"):
                file_head.append(line)  # $DATA
                line = file.readline()
                file_head.append(line)  # comment line
                line = file.readline()
                if not line.strip().upper().startswith("C1"):
                    raise ValueError(f"Symmetry group {line.strip()} is not supported")
                file_head.append(line)  # symmetry group (e.g. C1)
                break
            file_head.append(line)
        while True:
            line = file.readline()
            if line.strip().upper().startswith("$END"):
                file_tail.append(line)
                break
        while True:
            line = file.readline()
            if not line:
                break
            file_tail.append(line)
    return "".join(file_head), "".join(file_tail)


def gms_inp_parser(filepath: str | Path) -> Molecule:
    filepath = validate_filepath(filepath)
    with zopen(filepath, mode="rt", encoding="utf-8") as file:
        content = file.read()
    mol = from_gms_inp_str(content)
    mol.name = filepath.stem

    return mol
